package nz.ac.massey.javaecs;
/**
 * Entity Manager
 * This class manages entity allocations; keeping a list
 * of all unassigned entities values. 
 * Additionally, handles setting the registered component
 * flags for an entity.
 * 
 * Contributors:
 * Brychan Dempsey - brychand@hotmail.com
 * 
 */

import java.util.BitSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.ArrayList;

// Define the manager classes internally - should be moved to seperate source files as appropriate
/**
 * Manages data from the perspective of the entity.
 * <p>
 * I.e. Controls adding and removing entities, and registration and deregistration of components.
 */
class EntityManager{
    Queue<Integer> unusedEntities;
    List<BitSet> entityRegistrations;
    int currentSize;

    public EntityManager(){
        currentSize = 1024;
        unusedEntities = new LinkedList<>();
        entityRegistrations = new ArrayList<>();

        for (int i = 0; i < 1024; i++) {
            unusedEntities.add(i);
            entityRegistrations.add(new BitSet());
        }
    }

    public EntityManager(int maxEntities){
        currentSize = maxEntities;
        unusedEntities = new LinkedList<>();
        entityRegistrations = new ArrayList<>();
        
        for (int i = 0; i < maxEntities; i++) {
            unusedEntities.add(i);
            entityRegistrations.add(new BitSet());
        }
    }


    public Integer addEntity(){
        if (unusedEntities.size() == 0){
            System.err.println("No available space to create a new entity");
            return -1;
        }
        return unusedEntities.remove();
    }

    public void removeEntity(int entity){
        unusedEntities.add(entity);
        entityRegistrations.get(entity).clear();
    }

    public void registerComponent(int component, int entity){
        if (entity >= currentSize){
            System.err.println("Attempted to assign a component to non-existent entity: " + entity);
            return;
        }
        entityRegistrations.get(entity).set(component);
    }

    public void unregisterComponent(int component, int entity){
        entityRegistrations.get(entity).clear(component);
    }

    public BitSet getRegistrations(int entity){
        return entityRegistrations.get(entity);
    }

    public void setRegistrations(int entity, BitSet registrations){
        entityRegistrations.set(entity, registrations);
    }

    public boolean resize(int newSize, SystemManager systemManager, ComponentManager componentManager){
        if (newSize < currentSize - unusedEntities.size()){
            System.err.println("Attempted to resize the maximum entity count to a number smaller than the current assigned entity count.");
            return false;
        }
        else if (newSize == currentSize){
            System.err.println("Attempted to set the newSize to the current size");
            return true;
        }
        else{
            // Consistency should be maintained.
            // This is computationally expensive; we must re-order every assigned entity above newSize, if the newSize is smaller
            if (newSize < currentSize){
                List<Integer> outOfBounds = new ArrayList<>();
                for (int i = newSize; i < currentSize; i++) {
                    if (!unusedEntities.remove(i)){
                        // Could not remove element, as it didn't exist (already assigned).
                        // must find it and reassign it a new in-bounds value
                        outOfBounds.add(i);
                    }
                }
                for (Integer integer : outOfBounds) {
                    int newPos = addEntity();
                    setRegistrations(newPos, getRegistrations(integer));
                    // Invoke an update in the SystemManager
                    systemManager.entityDestroyed(integer);
                    systemManager.entitySignatureChanged(newPos, getRegistrations(newPos));
                    // Invoke the change in the components
                    componentManager.moveComponentData(integer, newPos);
                    componentManager.entityDestroyed(integer);
                }
                for (int i = newSize; i < currentSize; i++) {
                    // Remove out-of-bounds data
                    entityRegistrations.remove(newSize);
                }
            }
            else{
                // Init unassigned values
                for (int i = currentSize; i < newSize; i++) {
                    unusedEntities.add(i);
                    entityRegistrations.add(new BitSet());
                }
            }
            // Finally, set the current size
            currentSize = newSize;
            return true;
        }
    }
}